#include "unittest.h"
#include "../processXml.h"
#include "../../shared/xmlBase.h"
#include <wx/filefn.h>
#include <wx/intl.h>
#include "../../shared/standardPaths.h"

using namespace FreeFileSync;


TEST(XmlInputOutput, GUI)
{
    //cleanup
    wxString testXml = wxT("test語.xml");
    if (wxFileExists(testXml))
        wxRemoveFile(testXml);

    ASSERT_FALSE(wxFileExists(testXml));

    //load structure with gui settings "guiCfg"
    xmlAccess::XmlGuiConfig guiCfg;
    guiCfg.mainCfg.compareVar = CMP_BY_TIME_SIZE;
    guiCfg.mainCfg.syncConfiguration.exLeftSideOnly  = FreeFileSync::SYNC_DIR_LEFT;
    guiCfg.mainCfg.syncConfiguration.exRightSideOnly = FreeFileSync::SYNC_DIR_NONE;
    guiCfg.mainCfg.syncConfiguration.leftNewer       = FreeFileSync::SYNC_DIR_LEFT;
    guiCfg.mainCfg.syncConfiguration.rightNewer      = FreeFileSync::SYNC_DIR_RIGHT;
    guiCfg.mainCfg.syncConfiguration.different       = FreeFileSync::SYNC_DIR_RIGHT;
    guiCfg.mainCfg.filterIsActive = false;
    guiCfg.mainCfg.includeFilter = wxT("include語\\");
    guiCfg.mainCfg.excludeFilter = wxT("語exclude\\8");
    guiCfg.mainCfg.handleDeletion = FreeFileSync::MOVE_TO_CUSTOM_DIRECTORY;
    guiCfg.ignoreErrors  = false;

    FolderPair pair;
    pair.leftDirectory  = wxT("&left1語");
    pair.rightDirectory = wxT("&right1語");
    guiCfg.directoryPairs.push_back(pair);

    pair.leftDirectory  = wxT("国語1");
    pair.rightDirectory = wxT("国語2");
    guiCfg.directoryPairs.push_back(pair);

    guiCfg.hideFilteredElements = false;

    //write XML
    ASSERT_NO_THROW(writeGuiConfig(testXml, guiCfg));

    ASSERT_TRUE(xmlAccess::getXmlType(testXml) == xmlAccess::XML_GUI_CONFIG);

    //read XML
    xmlAccess::XmlGuiConfig guiCfg2;
    ASSERT_NO_THROW(guiCfg2 = xmlAccess::readGuiConfig(testXml));

    //test if retrieved file is equal created file
    ASSERT_TRUE(guiCfg == guiCfg2);

    //clean up
    wxRemoveFile(testXml);
}


TEST(XmlInputOutput, Batch)
{
    //cleanup
    wxString testXml = wxT("test語.xml");
    if (wxFileExists(testXml))
        wxRemoveFile(testXml);

    ASSERT_FALSE(wxFileExists(testXml));

    //load structure with batch settings "batchCfg"
    xmlAccess::XmlBatchConfig batchCfg;
    batchCfg.mainCfg.compareVar = CMP_BY_TIME_SIZE;
    batchCfg.mainCfg.syncConfiguration.exLeftSideOnly  = FreeFileSync::SYNC_DIR_LEFT;
    batchCfg.mainCfg.syncConfiguration.exRightSideOnly = FreeFileSync::SYNC_DIR_NONE;
    batchCfg.mainCfg.syncConfiguration.leftNewer       = FreeFileSync::SYNC_DIR_LEFT;
    batchCfg.mainCfg.syncConfiguration.rightNewer      = FreeFileSync::SYNC_DIR_RIGHT;
    batchCfg.mainCfg.syncConfiguration.different       = FreeFileSync::SYNC_DIR_RIGHT;
    batchCfg.mainCfg.filterIsActive = false;
    batchCfg.mainCfg.includeFilter = wxT("include語\\");
    batchCfg.mainCfg.excludeFilter = wxT("語exclude\\8");
    batchCfg.mainCfg.handleDeletion = FreeFileSync::MOVE_TO_CUSTOM_DIRECTORY;
    batchCfg.handleError = xmlAccess::ON_ERROR_IGNORE;

    FolderPair pair;
    pair.leftDirectory  = wxT("&left1語");
    pair.rightDirectory = wxT("&right1語");
    batchCfg.directoryPairs.push_back(pair);

    pair.leftDirectory  = wxT("国語1");
    pair.rightDirectory = wxT("国語2");
    batchCfg.directoryPairs.push_back(pair);

    batchCfg.silent = true;

    //write XML
    ASSERT_NO_THROW(writeBatchConfig(testXml, batchCfg));

    ASSERT_TRUE(xmlAccess::getXmlType(testXml) == xmlAccess::XML_BATCH_CONFIG);

    //read XML
    xmlAccess::XmlBatchConfig batchCfg2;
    ASSERT_NO_THROW(batchCfg2 = xmlAccess::readBatchConfig(testXml));

    //test if retrieved file is equal created file
    ASSERT_TRUE(batchCfg.directoryPairs == batchCfg2.directoryPairs);
    ASSERT_TRUE(batchCfg.mainCfg == batchCfg2.mainCfg);
    ASSERT_EQ(batchCfg.handleError, batchCfg2.handleError);

    ASSERT_EQ(batchCfg.silent, batchCfg2.silent);

    //clean up
    wxRemoveFile(testXml);
}


inline
bool colAttrIsEqual(const xmlAccess::ColumnAttrib& a, const xmlAccess::ColumnAttrib& b)
{
    return (a.type     == b.type     &&
            a.visible  == b.visible  &&
            a.position == b.position &&
            a.width    == b.width);
}


TEST(XmlInputOutput, GlobalSettings)
{
    //load structure with settings
    xmlAccess::XmlGlobalSettings globalCfg;

    globalCfg.programLanguage         = wxLANGUAGE_ICELANDIC;
    globalCfg.fileTimeTolerance       = 123;
    globalCfg.warnings.warningDependentFolders = false;

    globalCfg.gui.widthNotMaximized    = 1111;
    globalCfg.gui.heightNotMaximized   = 2222;
    globalCfg.gui.posXNotMaximized     = 3333;
    globalCfg.gui.posYNotMaximized     = 4444;
    globalCfg.gui.isMaximized          = true;

    for (int i = 0; i < 4; ++i)
    {
        xmlAccess::ColumnAttrib newEntry;

        newEntry.type     = xmlAccess::ColumnTypes(i % 4);
        newEntry.visible  = true;
        newEntry.position = i;
        newEntry.width    = i + 1000;
        globalCfg.gui.columnAttribLeft.push_back(newEntry);

        newEntry.type     = xmlAccess::ColumnTypes(i % 4);
        newEntry.visible  = false;
        newEntry.position = i;
        newEntry.width    = i + 2000;
        globalCfg.gui.columnAttribRight.push_back(newEntry);
    }

    //write XML
    ASSERT_NO_THROW(writeGlobalSettings(globalCfg));

    ASSERT_TRUE(xmlAccess::getXmlType(FreeFileSync::getGlobalConfigFile()) == xmlAccess::XML_GLOBAL_SETTINGS);

    //read XML
    xmlAccess::XmlGlobalSettings globalCfg2;
    ASSERT_NO_THROW(globalCfg2 = xmlAccess::readGlobalSettings());

    //test if retrieved file is equal created file
    ASSERT_EQ(globalCfg.programLanguage, globalCfg2.programLanguage);
    ASSERT_EQ(globalCfg.fileTimeTolerance, globalCfg2.fileTimeTolerance);
    ASSERT_EQ(globalCfg.warnings.warningDependentFolders, globalCfg2.warnings.warningDependentFolders);

    ASSERT_EQ(globalCfg.gui.widthNotMaximized, globalCfg2.gui.widthNotMaximized);
    ASSERT_EQ(globalCfg.gui.heightNotMaximized, globalCfg2.gui.heightNotMaximized);
    ASSERT_EQ(globalCfg.gui.posXNotMaximized, globalCfg2.gui.posXNotMaximized);
    ASSERT_EQ(globalCfg.gui.posYNotMaximized, globalCfg2.gui.posYNotMaximized);
    ASSERT_EQ(globalCfg.gui.isMaximized, globalCfg2.gui.isMaximized);
    ASSERT_EQ(globalCfg.gui.columnAttribLeft.size(), globalCfg2.gui.columnAttribLeft.size());
    ASSERT_EQ(globalCfg.gui.columnAttribRight.size(), globalCfg2.gui.columnAttribRight.size());
    ASSERT_TRUE(equal(globalCfg.gui.columnAttribLeft.begin(), globalCfg.gui.columnAttribLeft.end(), globalCfg2.gui.columnAttribLeft.begin(), colAttrIsEqual));
    ASSERT_TRUE(equal(globalCfg.gui.columnAttribRight.begin(), globalCfg.gui.columnAttribRight.end(), globalCfg2.gui.columnAttribRight.begin(), colAttrIsEqual));
}
