#include "unittest.h"
#include "../resources.h"
#include "../../shared/fileHandling.h"
#include "../multithreading.h"
#include "../../algorithm.h"

using namespace FreeFileSync;


TEST(Formatting, DirectoryNames)
{
    ASSERT_PRED_FORMAT2(assertEqual, FreeFileSync::getFormattedDirectoryName(wxT("  c: ")).c_str(), wxT("c:\\"));
    ASSERT_PRED_FORMAT2(assertEqual, FreeFileSync::getFormattedDirectoryName(wxT("  c:\\test\\ ")).c_str(), wxT("c:\\test\\"));
    ASSERT_PRED_FORMAT2(assertEqual, FreeFileSync::getFormattedDirectoryName(wxT("  \\abc\\ ")).c_str(), wxT("\\abc\\"));
    ASSERT_PRED_FORMAT2(assertEqual, FreeFileSync::getFormattedDirectoryName(wxT("  \\a\\b\\c ")).c_str(), wxT("\\a\\b\\c\\"));
    ASSERT_PRED_FORMAT2(assertEqual, FreeFileSync::getFormattedDirectoryName(wxT("  \\a\\b\\c d  ")).c_str(), wxT("\\a\\b\\c d\\"));
    ASSERT_PRED_FORMAT2(assertEqual, FreeFileSync::getFormattedDirectoryName(wxT("  \\a b/b\\c d  ")).c_str(), wxT("\\a b/b\\c d\\"));
}


//predicate-formatter for asserting that two vector<Zstring> are equal
testing::AssertionResult assertEqual(const char* a_expr, const char* b_expr, const std::vector<Zstring>& a, const std::vector<Zstring>& b)
{
    if (a == b)
        return testing::AssertionSuccess();

    testing::Message msg;
    msg << a_expr << " = " << '\n';
    for (std::vector<Zstring>::const_iterator i = a.begin(); i != a.end(); ++i)
        msg << i->c_str() << '\n';

    msg << " and " << b_expr << " = " << '\n';
    for (std::vector<Zstring>::const_iterator i = b.begin(); i != b.end(); ++i)
        msg << i->c_str() << '\n';

    msg << " not equal!";

    return testing::AssertionFailure(msg);
}

std::vector<Zstring> compoundStringToFilter(const Zstring& filterString);

TEST(Formatting, FilterString)
{
    std::vector<Zstring> filterActual = compoundStringToFilter(wxT("   A\n   b;; c;\n\n;\n d    ;E    f  \n;& ;g\\"));
    std::vector<Zstring> filterExpected;
    filterExpected.push_back(wxT("A"));
    filterExpected.push_back(wxT("b"));
    filterExpected.push_back(wxT("c"));
    filterExpected.push_back(wxT("d"));
    filterExpected.push_back(wxT("E    f"));
    filterExpected.push_back(wxT("&"));
    filterExpected.push_back(wxT("g\\"));
    ASSERT_PRED_FORMAT2(assertEqual, filterActual, filterExpected);

    filterActual = compoundStringToFilter(wxT("   ; ;\n a"));
    filterExpected.clear();
    filterExpected.push_back(wxT("a"));
    ASSERT_PRED_FORMAT2(assertEqual, filterActual, filterExpected);

    filterActual = compoundStringToFilter(wxT("   ; ;\n       ;\n\n"));
    filterExpected.clear();
    ASSERT_PRED_FORMAT2(assertEqual, filterActual, filterExpected);
}


TEST(Formatting, FileSize)
{
    wxString sizeActual = FreeFileSync::formatFilesizeToShortString(807.0);
    wxString sizeExpected = wxString(wxT("807")) + _(" Byte");
    ASSERT_PRED_FORMAT2(assertEqual, sizeActual, sizeExpected);

    sizeActual = FreeFileSync::formatFilesizeToShortString(315621376.0);
    sizeExpected = wxString::Format(wxT("%i"), 301) + _(" MB");
    ASSERT_PRED_FORMAT2(assertEqual, sizeActual, sizeExpected);

    sizeActual = FreeFileSync::formatFilesizeToShortString(48954.0);
    sizeExpected = wxString(wxT("47.8")) + _(" kB");
    ASSERT_PRED_FORMAT2(assertEqual, sizeActual, sizeExpected);

    sizeActual = FreeFileSync::formatFilesizeToShortString(35032904.0);
    sizeExpected = wxString(wxT("33.4")) + _(" MB");
    ASSERT_PRED_FORMAT2(assertEqual, sizeActual, sizeExpected);

    sizeActual = FreeFileSync::formatFilesizeToShortString(3084582912.0);
    sizeExpected = wxString(wxT("2.87")) + _(" GB");
    ASSERT_PRED_FORMAT2(assertEqual, sizeActual, sizeExpected);

    sizeActual = FreeFileSync::formatFilesizeToShortString(1001.0);
    sizeExpected = wxString(wxT("0.97")) + _(" kB");
    ASSERT_PRED_FORMAT2(assertEqual, sizeActual, sizeExpected);
}


//handle execution of a method while updating the UI
class UpdateDummy : public UpdateWhileExecuting
{
private:
    void longRunner() {} //virtual impl.
};


TEST(Miscellaneous, Standard)
{
    //for Windows Recycle Bin should always be accessible
#ifdef FFS_WIN
    ASSERT_TRUE(FreeFileSync::recycleBinExists());
#endif

    //test if instantiation and deletion of multithreading class works correctly (without hang)
    for (int i = 0; i < 50; ++i)
    {
        UpdateDummy d;
    }

    //test if multiple executions work (without hang)
    UpdateDummy d;
    DummyStatusUpdater dummy;
    for (int i = 0; i < 1000; ++i)
    {
        d.waitUntilReady();
        d.execute(&dummy);
    }
}
